"""Broadcast commands sent from the device.

These messages represent commands sent to the Insteon Modem where
the device is the controller and the modem is the responder in the
All-Link database.

Broadcast commands are device state change notifications from the
device to any responder.
"""
