"""Device state definitions."""
DIMMABLE_LIGHT = "dimmable_light"
DIMMABLE_FAN = "dimmable_fan"
DIMMABLE_OUTLET = "dimmable_outlet"
ON_OFF_SWITCH = "on_off_switch"
RELAY = "RELAY"
ON_OFF_OUTLET_TOP = "on_off_outlet_top"
ON_OFF_OUTLET_BOTTOM = "on_off_outlet_bottom"
LOW_BATTERY = "low_battery"
OPEN_CLOSE_SENSOR = "open_close_sensor"
LIGHT_SENSOR = "light_sensor"
LOAD_SENSOR = "load_sensor"
LEAK_SENSOR_DRY = "leak_sensor_dry"
LEAK_SENSOR_WET = "leak_sensor_wet"
DOOR_SENSOR = "door_sensor"
MOTION_SENSOR = "motion_sensor"
SMOKE_SENSOR = "smoke_sensor"
CO_SENSOR = "co_sensor"
TEST_SENSOR = "test_sensor"
NEW_SENSOR = "new_sensor"
HEARTBEAT = "heartbeat"
SENSOR_MALFUNCTION = "sensor_malfunction"
COVER = "cover"
COOL_SET_POINT = "cool_set_point"
HEAT_SET_POINT = "heat_set_point"
HUMIDITY_HIGH = "humidity_high"
HUMIDITY_LOW = "humidity_low"
SYSTEM_MODE = "system_mode"
FAN_MODE = "fan_mode"
TEMPERATURE = "temperature"
HUMIDITY = "humidity"
COOLING = "cooling"
HEATING = "heating"
HUMIDIFYING = "humidifying"
DEHUMIDIFYING = "dehumidifying"

ON_OFF_SWITCH_MAIN = f"{ON_OFF_SWITCH}_main"
ON_OFF_SWITCH_A = f"{ON_OFF_SWITCH}_a"
ON_OFF_SWITCH_B = f"{ON_OFF_SWITCH}_b"
ON_OFF_SWITCH_C = f"{ON_OFF_SWITCH}_c"
ON_OFF_SWITCH_D = f"{ON_OFF_SWITCH}_d"
ON_OFF_SWITCH_E = f"{ON_OFF_SWITCH}_e"
ON_OFF_SWITCH_F = f"{ON_OFF_SWITCH}_f"
ON_OFF_SWITCH_G = f"{ON_OFF_SWITCH}_g"
ON_OFF_SWITCH_H = f"{ON_OFF_SWITCH}_h"

DIMMABLE_LIGHT_MAIN = f"{DIMMABLE_LIGHT}_main"
